mod api;
mod cli;
mod config;
pub mod db;
mod error;
mod repo;

pub use config::{Config, DbConfig, FsConfig};
pub use error::{Result, ServerError};

use std::{io, path::PathBuf, sync::Arc};

use axum::Router;
use tower_http::trace::TraceLayer;

use clap::Parser;
use sea_orm_migration::MigratorTrait;
use tokio::runtime;
use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt};

pub const ANY_ARCH: &'static str = "any";

#[derive(Clone)]
pub struct Global {
    config: crate::config::Config,
    repo: repo::Handle,
    db: sea_orm::DbConn,
}

fn main() -> crate::Result<()> {
    let rt = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()
        .unwrap();
    let handle = rt.handle();

    let cli = cli::Cli::parse();
    let global = setup(handle, cli.config_file)?;

    handle.block_on(run(global))
}

fn setup(rt: &runtime::Handle, config_file: PathBuf) -> crate::Result<Global> {
    let config: Config = Config::figment(config_file)
        .extract()
        .inspect_err(|e| tracing::error!("{}", e))?;

    tracing_subscriber::registry()
        .with(tracing_subscriber::EnvFilter::new(config.log_level.clone()))
        .with(tracing_subscriber::fmt::layer())
        .init();

    tracing::info!("Connecting to database");
    let db = rt.block_on(crate::db::connect(&config.db))?;
    rt.block_on(crate::db::Migrator::up(&db, None))?;

    let repo = match &config.fs {
        FsConfig::Local { data_dir } => {
            crate::repo::Handle::start(
                data_dir.join("repos"),
                db.clone(),
                rt.clone(),
                config.pkg_workers,
            )?
            //rt.block_on(crate::repo::RepoMgr::new(
            //    data_dir.join("repos"),
            //    db.clone(),
            //))?
            //RepoHandle::start(data_dir.join("repos"), db.clone(), config.pkg_workers, rt.clone())?
        }
    };
    //let mgr = Arc::new(mgr);
    //
    //for _ in 0..config.pkg_workers {
    //    let clone = Arc::clone(&mgr);
    //
    //    rt.spawn(async move { clone.pkg_parse_task().await });
    //}

    Ok(Global {
        config: config.clone(),
        repo,
        db,
    })
}

async fn run(global: Global) -> crate::Result<()> {
    let domain: String = format!("{}:{}", &global.config.domain, global.config.port)
        .parse()
        .unwrap();
    let listener = tokio::net::TcpListener::bind(domain).await?;

    // build our application with a single route
    let app = Router::new()
        .nest("/api", crate::api::router())
        .merge(crate::repo::router(&global.config.api_key))
        .with_state(global)
        .layer(TraceLayer::new_for_http());
    // run it with hyper on localhost:3000
    Ok(axum::serve(listener, app.into_make_service())
        .await
        .map_err(|err| io::Error::new(io::ErrorKind::Other, err))?)
}
